import sys
from collections import Counter, defaultdict
from typing import Collection, List, Set, Dict, Union, Counter as TCounter, Optional

from pb_amarder import Progress
from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.ixps import PeeringDB
from traceutils2.radix.ip2as import IP2AS
from traceutils2.utils.utils import max_num, peek

from bdrmapit.algorithm import debug
from bdrmapit.algorithm.debug import DebugMixin
from bdrmapit.algorithm.firsthopmixin import FirstHopMixin
from bdrmapit.algorithm.helpersmixin import HelpersMixin
from bdrmapit.algorithm.lasthopsmixin import LastHopsMixin
from bdrmapit.algorithm.regexmixin import RegexMixin
from bdrmapit.algorithm.utypes import HIDDEN_NOINTER, SINGLE_SUCC_4, ALLPEER_SUCC, VOTE_SINGLE, \
    VOTE_TIE, HIDDEN_INTER
from bdrmapit.algorithm.vrfmixin import VRFMixin
from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Router, Interface
from bdrmapit.vrf.vrfedge import VRFEdge


class Bdrmapit(FirstHopMixin, LastHopsMixin, VRFMixin, RegexMixin, DebugMixin, HelpersMixin):

    def __init__(self, graph: Graph, as2org: AS2Org, bgp: BGP, ixpasns=None, strict=False, skipua=False, hidden_reverse=True, norelpeer: Set[int]=None, interfaces: Collection[Interface]=None):
        self.graph = graph
        self.as2org = as2org
        self.bgp = bgp
        # self.peeringdb = peeringdb
        self.rupdates = Updates()
        self.iupdates = Updates()
        self.caches = Updates()
        self.routers_mpls: List[Router] = []
        self.lasthops: List[Router] = []
        self.routers_succ: List[Router] = []
        self.routers_vrf: List[Router] = []
        if interfaces is None:
            routers = graph.routers.values()
        else:
            routers = (iface.router for iface in interfaces)
        for router in routers:
            if router.succ:
                if router.vrf:
                    self.routers_vrf.append(router)
                else:
                    self.routers_succ.append(router)
            else:
                self.lasthops.append(router)
        self.routers_vrf = sorted(self.routers_vrf, key=self.sort_vrf)
        if interfaces is None:
            interfaces = graph.interfaces.values()
        self.interfaces_pred: List[Interface] = [i for i in interfaces if i.pred]
        self.previous_updates = []
        self.strict = strict
        self.skipua = skipua
        self.hidden_reverse = hidden_reverse
        self.norelpeer = norelpeer
        self.ixpasns = {} if ixpasns is None else ixpasns

    def peeringdb_ixpasns(self, peeringdb, ip2as: IP2AS):
        if isinstance(peeringdb, str):
            peeringdb = PeeringDB(peeringdb)
        self.ixpasns = defaultdict(set)
        for addr, asn in peeringdb.addrs.items():
            self.ixpasns[ip2as[addr]].add(asn)
        self.ixpasns.default_factory = None

    def router_heuristics(self, router: Router, isucc: Interface, origins: Set[int], iasns: TCounter[int], use_third=True, use_update=True):
        rsucc: Router = isucc.router  # subsequent router
        rsucc_asn = self.rupdates.asn(rsucc)  # subsequent router AS annotation
        iupdate = self.iupdates[isucc]  # update for subsequent interface (interface annotation)
        if debug.DEBUG:
            print(iupdate)

        # If subsequent interface is an IXP interface, use interface AS
        if isucc.asn <= -100:
            # ixp_succs = sum(1 for s in router.succ if s.asn == isucc.asn)
            # if ixp_succs >= 3 or ixp_succs >= len(router.succ) * .5:
            ixpasns = self.ixpasns.get(isucc.asn)
            if ixpasns:
                if debug.DEBUG: print('IXP ASNs: {}'.format(len(ixpasns)))
                overlap = iasns.keys() & ixpasns
                if len(overlap) == 1:
                    return peek(overlap)
            return -1

        # If subsequent interface AS has no known origin, use subsequent router AS
        if isucc.asn == 0:
            return rsucc_asn if not self.skipua else -1

        if use_update and iupdate and self.as2org[rsucc_asn] == isucc.org:
            succ_asn = iupdate.asn
            succ_org = iupdate.org
            if succ_asn <= 0:
                succ_asn = isucc.asn
                succ_org = isucc.org
            elif False and succ_asn not in origins and len(router.dests & self.bgp.cone[isucc.asn]) > len(router.dests & self.bgp.cone[succ_asn]):
                if debug.DEBUG:
                    print(len(router.dests), len(router.dests & self.bgp.cone[isucc.asn]), len(router.dests & self.bgp.cone[succ_asn]))
                succ_asn = isucc.asn
                succ_org = isucc.org
            # elif isucc.asn in router.dests and all(iasn <= 0 for iasn in origins):
            #     succ_asn = isucc.asn
            #     succ_org = isucc.org
        else:
            succ_asn = isucc.asn
            succ_org = isucc.org
        if debug.DEBUG: print('\tASN={}, RASN={}, IUpdate={} VRF={}'.format(isucc.asn, rsucc_asn, succ_asn, router.vrf))

        # Third party stuff
        if use_third:
            third = False
            if any(iasn > 0 for iasn in origins) and not any(isucc.org == self.as2org[iasn] for iasn in origins):
                # If here, subsequent interface is not in IR origin ASes
                if rsucc_asn > 0:
                    # If here, the subsequent router has an AS annotation
                    rsucc_org = self.as2org[rsucc_asn]
                    if rsucc_org != succ_org and not any(succ_org == self.as2org[iasn] for iasn in origins):
                        # If here, subsequent router AS is different from the subsequent interface AS
                        if debug.DEBUG:
                            print('\tThird party: Router={}, RASN={}'.format(rsucc.name, rsucc_asn))
                            print('\tAny origin-router rels? {}'.format(rsucc_asn in origins or self.any_rels(rsucc_asn, origins)))
                        if rsucc_asn in origins or self.any_rels(rsucc_asn, origins):
                            # If here, some origin AS matches the subsequent router's AS annotation
                            # Or, some origin AS directly interconnects with the subsequent routers AS annotation

                            # Number of destination in subsequent interface origin AS customer cone
                            s_conesize = len(router.dests & self.bgp.cone[succ_asn])
                            # Number of destination in subsequent router AS annotation customer cone
                            r_conesize = len(router.dests & self.bgp.cone[rsucc_asn])
                            if debug.DEBUG:
                                if len(router.dests) <= 5:
                                    print('\tISUCC in Dests: {} in {}'.format(succ_asn, router.dests))
                                else:
                                    print('\tISUCC not in Dests: {}'.format(succ_asn not in router.dests))
                                print('\t{} < {}'.format(s_conesize, r_conesize))
                            if succ_asn not in router.dests:
                                # If here, the subsequent AS is not in the router's destination ASes
                                if s_conesize <= r_conesize:
                                    # If here, at least as many destinations are in the router AS annotation cone than in the subsequent interface origin AS cone
                                    third = True
                            elif not self.any_rels(succ_asn, origins) and self.bgp.rel(succ_asn, rsucc_asn):
                                # If here, none of the origins connect to the subsequent interface origin AS,
                                # and the subsequent interface AS has a relationship with the router AS annotation.
                                third = True

                # When there is no relationship between router ASes and subsequent interface AS,
                # check if relationship between router ASes and subsequent router AS when they are the same org
                if succ_org == self.as2org[rsucc_asn]:
                    if not any(self.bgp.rel(iasn, succ_asn) for iasn in iasns):
                        if any(self.bgp.rel(iasn, rsucc_asn) for iasn in iasns):
                            if debug.DEBUG: print('Testing')
                            # return rsucc_asn
                            third = True
            if third:
                # Third party was detected!
                rsucc_cone = self.bgp.cone[rsucc_asn]  # subsequent router AS annotation customer cone
                if debug.DEBUG:
                    if len(router.dests) <= 5:
                        print('\tDests: {}'.format(router.dests))
                    if len(rsucc_cone) <= 5:
                        print('\tCone: {}'.format(rsucc_cone))
                if all(dasn == rsucc_asn or dasn in rsucc_cone for dasn in router.dests):
                    # If here, all destination ASes are in the customer cone of the subsequent router's AS annotation
                    return rsucc_asn
                if debug.DEBUG:
                    for origin in origins:
                        print('\tOrigin {}: RSUCC overlap {} ? SUCC overlap {}'.format(origin, len(router.dests & rsucc_cone), len(router.dests & self.bgp.cone[origin])))
                # Otherwise, ignore vote
                return -1

        # TODO: Figure out something better to do here
        if succ_asn <= 0 or (rsucc_asn > 0 and self.as2org[rsucc_asn] != isucc.org):
            if debug.DEBUG:
                if succ_asn != isucc.asn:
                    print('ugh')
            succ_asn = isucc.asn
        return succ_asn

    def hidden_asn(self, iasns: TCounter[int], asn: int, utype: int, votes: Dict[int, int]):
        """
        Look for hidden AS between the interface AS and the selected AS.
        """
        intasn: Optional[int] = None
        # First look for customers of interface AS who are providers of selected AS
        intersection: Set[int] = self.multi_customers(iasns) & self.bgp.providers[asn]
        # Only use if the intersection is definitive, i.e., a single AS
        if len(intersection) == 1:
            intasn = peek(intersection)
            if debug.DEBUG: print('Hidden: {}'.format(intasn))

        # If there is no intersection, check for provider of interface AS who is customer of selected AS
        elif not intersection and self.hidden_reverse:
            intersection = self.multi_providers(iasns) & self.bgp.customers[asn]
            if len(intersection) == 1:
                intasn = peek(intersection)
                if debug.DEBUG: print('Hidden Reversed: {}'.format(intasn))

        # If a hidden AS was selected
        if intasn is not None:
            interorg = self.as2org[intasn]
            # If the hidden AS is a sibling of an AS which received a vote (interface or subsequent AS),
            #  use the selected AS. I think it suggests the selected AS actually has a relationship.
            if interorg in {self.as2org[vasn] for vasn in votes}:
                return asn, HIDDEN_NOINTER + utype
            return intasn, HIDDEN_INTER + utype

        if debug.DEBUG: print('Missing: {}-{}'.format(iasns, asn))
        if self.strict:
            return max(iasns, key=lambda x: (votes[x], -self.bgp.conesize[x], x)), HIDDEN_NOINTER + utype
        else:
            return asn, HIDDEN_NOINTER + utype

    def isnorelpeer(self, iasn):
        return bool(self.norelpeer and iasn in self.norelpeer)

    def count_votes_test(self, router):
        if isinstance(router, str):
            router = self.graph.routers[router]
        iasns: TCounter[int] = Counter(interface.asn for interface in router.interfaces if interface.asn > 0)
        succs_init = Counter()
        succs = Counter()  # Subsequent interface vote recorder
        sasn_origins = defaultdict(set)  # Origins seen prior to subsequent interfaces
        succ_map = defaultdict(list)
        self.count_votes(router, iasns, succs_init, succs, sasn_origins, succ_map)
        return succs

    def count_votes(self, router, iasns, succs_init, succs, sasn_origins, succ_map, remap=True):
        for isucc in router.succ:
            origins = router.origins[isucc]  # Origins seen prior to subsequent interface
            if debug.DEBUG: print('Succ={}, ASN={}, Origins={} RSucc={}'.format(isucc.addr, isucc.asn, origins, isucc.router.name))
            if remap:
                succ_asn_init = self.router_heuristics(router, isucc, origins, iasns, use_third=False, use_update=False)  # AS vote for the subsequent interface
                if debug.DEBUG:
                    print('\tHeuristic Init: {}'.format(succ_asn_init))
                # If vote is useful
                if succ_asn_init > 0:
                    succs_init[succ_asn_init] += 1
            succ_asn = self.router_heuristics(router, isucc, origins, iasns)  # AS vote for the subsequent interface
            if debug.DEBUG:
                print('\tHeuristic: {}'.format(succ_asn))
            if succ_asn > 0:
                succs[succ_asn] += 1  # record vote
                sasn_origins[succ_asn].update(origins)  # record origin ASes seen before interface
                succ_map[succ_asn].append(isucc)
        if debug.DEBUG: print('Succs: {}'.format(succs))

        if remap:
            if router in self.rupdates:
                for sasn, isuccs in succ_map.items():
                    if not all(isucc.router in self.rupdates for isucc in isuccs):
                        continue
                    rasns = Counter(self.rupdates[isucc.router].asn for isucc in isuccs)
                    rasn = max(rasns, key=lambda x: rasns[x])
                    if (len(isuccs) + succs.get(rasn, 0) >= 3) and rasns[rasn] >= len(isuccs) * 0.9:
                        if rasn != sasn:
                            if debug.DEBUG:
                                print('Remapping {} -> {} ({}/{}={})'.format(sasn, rasn, rasns[rasn], len(isuccs), rasns[rasn] / len(isuccs)))
                            succs[rasn] += succs[sasn]
                            del succs[sasn]
            if debug.DEBUG:
                print('Succs: {}'.format(succs))

    def annotate_router(self, router: Router, first=False, remap=False, **kwargs):
        isucc: Union[Interface, VRFEdge]
        utype: int = 0

        # Router origin ASes
        iasns: TCounter[int] = Counter(interface.asn for interface in router.interfaces if interface.asn > 0)
        if debug.DEBUG:
            if len(router.interfaces) <= 5:
                print('Interfaces: {}'.format(' '.join(i.addr for i in router.interfaces)))
            print('IASN: {}'.format(iasns))
            print('Edges={}, NH={}'.format(len(router.succ), router.nexthop))
            print('VRF={}'.format(router.vrf))

        # Use heuristics to determine link votes
        succs_init = Counter()
        succs = Counter()  # Subsequent interface vote recorder
        sasn_origins = defaultdict(set)  # Origins seen prior to subsequent interfaces
        succ_map = defaultdict(list)
        self.count_votes(router, iasns, succs_init, succs, sasn_origins, succ_map, remap=remap)
        # # For each subsequent interface
        # for isucc in router.succ:
        #     origins = router.origins[isucc]  # Origins seen prior to subsequent interface
        #     if debug.DEBUG: print('Succ={}, ASN={}, Origins={} RSucc={}'.format(isucc.addr, isucc.asn, origins, isucc.router.name))
        #     succ_asn_init = self.router_heuristics(router, isucc, origins, iasns, use_third=False, use_update=False)  # AS vote for the subsequent interface
        #     succ_asn = self.router_heuristics(router, isucc, origins, iasns)  # AS vote for the subsequent interface
        #     if debug.DEBUG:
        #         print('\tHeuristic: {}'.format(succ_asn))
        #         print('\tHeuristic Init: {}'.format(succ_asn_init))
        #     # If vote is useful
        #     if succ_asn_init > 0:
        #         succs_init[succ_asn_init] += 1
        #     if succ_asn > 0:
        #         succs[succ_asn] += 1  # record vote
        #         sasn_origins[succ_asn].update(origins)  # record origin ASes seen before interface
        #         succ_map[succ_asn].append(isucc)
        # if debug.DEBUG: print('Succs: {}'.format(succs))
        #
        # if router in self.rupdates:
        #     for sasn, isuccs in succ_map.items():
        #         if not all(isucc.router in self.rupdates for isucc in isuccs):
        #             continue
        #         rasns = Counter(self.rupdates[isucc.router].asn for isucc in isuccs)
        #         rasn = max(rasns, key=lambda x: rasns[x])
        #         if (len(isuccs) + succs.get(rasn, 0) >= 3) and rasns[rasn] >= len(isuccs) * 0.9:
        #             if rasn != sasn:
        #                 if debug.DEBUG:
        #                     print('Remapping {} -> {} ({}/{}={})'.format(sasn, rasn, rasns[rasn], len(isuccs), rasns[rasn] / len(isuccs)))
        #                 succs[rasn] += succs[sasn]
        #                 del succs[sasn]
        #
        # if debug.DEBUG:
        #     print('Succs: {}'.format(succs))

        sorigins = Counter(isucc.asn for isucc in router.succ)

        if len(succs_init) == 1 and len(router.succ) >= 3 and sum(succs_init.values()) == len(router.succ):
            return peek(succs_init), 20300

        # Create votes counter and add interface AS
        votes = succs + iasns
        if debug.DEBUG: print('Votes: {}'.format(votes))
        if not succs:
            # return -1, -1
            return self.annotate_lasthop(router)

        if False and all(interface.asn <= -100 for interface in router.interfaces):
            lasn, _ = self.annotate_lasthop(router)
            rasn = max(votes, key=lambda x: votes[x])
            numvotes = sum(votes.values())
            if numvotes < 4 or votes[rasn] / numvotes < .75:
                if lasn != rasn:
                    incone = rasn in self.bgp.cone[lasn]
                    inconenum = sum(1 for dest in router.dests if dest != lasn if dest in self.bgp.cone[lasn])
                    allincone = inconenum >= max(1., (len(router.dests)-1) * .9)
                    if debug.DEBUG:
                        print('Last Hop: {}'.format(lasn))
                        print('Most Votes: {}'.format(rasn))
                        print('{} in cone for {}? {}'.format(rasn, lasn, incone))
                        print('All in cone for {}? {} >= {}'.format(lasn, inconenum, len(router.dests) * 1))
                    if len(router.dests) > 2 and incone and allincone and self.bgp.conesize[lasn] < 100:
                        return lasn, 84321

        # if iasns and succs and any(sasn in iasns or sasn in self.multi_customers(sasn_origins[sasn]) for sasn in succs):
        #     maxrels = {}
        #     for asn in votes:
        #         rels = sum(bool(asn == dasn or self.bgp.rel(asn, dasn) or dasn in self.bgp.cone[asn]) for dasn in router.dests)
        #         maxrels[asn] = rels
        #     if debug.DEBUG:
        #         print('Max Rels: {}'.format(maxrels))
        #     if maxrels:
        #         asn, num = max(maxrels.items(), key=lambda x: x[1])
        #         if asn in iasns:
        #             if num > maxrels[max(succs, key=maxrels.__getitem__)] + 2:
        #                 return asn, 46

        if len(iasns) == 1 and len(succs) == 1:
            iasn = peek(iasns)
            sasn = peek(succs)
            if iasns[iasn] == succs[sasn]:
                if self.norelpeer and iasn in self.norelpeer and (self.bgp.peer_rel(iasn, sasn) or not self.bgp.rel(iasn, sasn)):
                # if self.bgp.peer_rel(iasn, sasn) or (self.norelpeer and iasn in self.norelpeer and not self.bgp.rel(iasn, sasn)):
                    return sasn, 5600

        # Deal specially with cases where there is only a single subsequent AS, or subsequent ORG
        # Multihomed exception
        if iasns and len(succs) == 1 or len({self.as2org[sasn] for sasn in succs}) == 1:
            # Subsequent AS
            sasn = peek(succs) if len(succs) == 1 else max(succs, key=lambda x: (self.bgp.conesize[x], -x))
            # Subsequent AS is not in link origin ASes and is a customer of a link origin AS
            if sasn not in sasn_origins[sasn]:
                if sasn in self.multi_customers(sasn_origins[sasn]):
                    if debug.DEBUG: print('Provider: {}->{}'.format(sasn_origins[sasn], sasn))
                    return sasn, utype + SINGLE_SUCC_4

        # Multiple Peers Exception
        # More than 1 subsequent AS and exactly one router origin AS
        if len(succs) > 1 and len(iasns) == 1:
            iasn = peek(iasns)
            # Origin AS is not also a subsequent AS
            if iasn not in succs:
                # All subsequent ASes are peers of the single origin AS
                peerrels = {self.as2org[sasn] for sasn in succs if self.bgp.peer_rel(iasn, sasn) or (self.norelpeer and iasn in self.norelpeer and not self.bgp.rel(iasn, sasn))}
                # numrels = sum(1 for sasn in succs if self.bgp.peer_rel(iasn, sasn) or (self.norelpeer and iasn in self.norelpeer and not self.bgp.rel(iasn, sasn)))
                numrels = len(peerrels)
                if debug.DEBUG: print('Peers: {} >= {}'.format(numrels, len(succs) * .85))
                if numrels >= len(succs) * .85:
                    max_vote = max(votes.values())
                    if debug.DEBUG:
                        print('All peers or norels: True')
                        print('IASN: {:,d}, Max Vote: {:,d}'.format(votes[iasn], max_vote))
                    if self.isnorelpeer(iasn):
                        # if (votes[iasn] > max_vote / 4 and len(succs) >= 2) or max_vote == 1:
                        #     asn, _ = self.annotate_lasthop(router)
                        #     if asn in succs:
                        #         return asn, 4532
                        if (votes[iasn] > max_vote / 4 and len(succs) >= 3) or max_vote == 1:
                            for x in succs:
                                if all(y in self.bgp.cone[x] for y in succs if y != x):
                                    return x, utype + ALLPEER_SUCC
                            if first:
                                return -1, utype + ALLPEER_SUCC
                            return iasn, utype + ALLPEER_SUCC
                    else:
                        # Make sure its votes are not dwarfed by subsequent AS
                        if votes[iasn] > max_vote / 2:
                            if first:
                                return -1, utype + ALLPEER_SUCC
                            # Select the router origin AS
                            return iasn, utype + ALLPEER_SUCC
                        if debug.DEBUG: print('{:,d} > {:.1f}'.format(votes[iasn], max_vote / 4))
                        if votes[iasn] > max_vote / 4 and sum(self.bgp.peer_rel(iasn, sasn) for sasn in succs) >= 2:
                            if first:
                                return -1, utype + ALLPEER_SUCC
                            return iasn, utype + ALLPEER_SUCC
                if len(succs) > 2:
                    numrels = sum(self.bgp.rel(iasn, sasn) for sasn in succs)
                    if debug.DEBUG: print('Rels: {} >= {}'.format(numrels, len(succs) * .9))
                    if numrels >= len(succs):
                        if votes[iasn] > max(votes.values()) / 2:
                            # Select the router origin AS
                            return iasn, utype + ALLPEER_SUCC
                        if votes[iasn] > max(votes.values()) / 4 and sum(
                                self.bgp.peer_rel(iasn, sasn) for sasn in succs) >= 2:
                            return iasn, utype + ALLPEER_SUCC
                        if self.norelpeer is not None and iasn in self.norelpeer and votes[iasn] > max(
                                votes.values()) / 4 and len(succs) >= 3:
                            return iasn, utype + ALLPEER_SUCC

        # Apply vote heuristics
        # asns -- maximum vote getters, often one AS, but will contain all tied ASes
        # Check if single AS accounts for at least 3/4 of votes
        othermax = max(votes, key=votes.__getitem__)
        if votes[othermax] >= sum(votes.values()) * .75:
            # If so, select that AS
            asns = [othermax]
        else:
            # Check if origin in succs
            # if iasns.keys() & succs.keys():
            #     succ_origins = set()
            #     for sasn, origins in sasn_origins.items():
            #         if sasn in origins:
            #             if debug.DEBUG: print('Succ {} in origins!'.format(sasn))
            #             succ_origins.add(sasn)
            #     if len(succ_origins) == 1 or len({self.as2org[orig] for orig in succ_origins}) == 1:
            #         return peek(succ_origins), 4567

            # Otherwise, find vote ASes with relationship to a router interface AS
            votes_rels = []
            for vasn in votes:
                if any(self.isnorelpeer(iasn) or vasn == iasn or self.bgp.rel(iasn, vasn) or self.as2org[iasn] == self.as2org[vasn] for iasn in iasns):
                    votes_rels.append(vasn)
            if debug.DEBUG: print('Vote Rels: {}'.format(votes_rels))

            # If only IR origin ASes remain, use all votes. Otherwise, use relationship ASes and origins
            if len(votes_rels) <= len(iasns):
                asns = max_num(votes, key=votes.__getitem__)
                if debug.DEBUG: print('ASNs: {}'.format(asns))
            else:
                asns = max_num(votes_rels, key=votes.__getitem__)

        # If single AS, select it
        if len(asns) == 1 and succs:
            asn = asns[0]
            utype += VOTE_SINGLE
        else:
            asn = None
            if len(iasns) == 1:
                iasn = peek(iasns)
                if self.norelpeer and iasn in self.norelpeer and iasn not in succs and len(succs) > 1 and all(votes[asn] == 1 for asn in asns):
                    return iasn, 24
            # Tiebreaker 1
            if asn is None and len(router.succ) == 1 and router.nexthop:
                isucc = peek(router.succ)  # single subsequent interface
                sasn = self.iupdates.asn(isucc)  # annotation for subsequent interface
                if len(router.interfaces) == 1 and sasn == -1:
                    rasn = router.interfaces[0].asn
                    if self.bgp.peer_rel(rasn, isucc.asn) or (self.isnorelpeer(rasn) and not self.bgp.rel(rasn, isucc.asn)):
                        return -1, 6000000
                # If annotation was used, is one of the tied ASes, and the subsequent interface has multiple incoming edges
                # if sasn in succs and sasn in asns and len(isucc.pred) > 1 and (True or not any(iasn > 0 and iasn in router.dests for iasn in sasn_origins[sasn])):
                if sasn in succs and sasn in asns and len(isucc.pred) > 1:
                    if debug.DEBUG: print('Pred Num: {}'.format(len(isucc.pred)))
                    asn = sasn  # select the subsequent interface annotation
                    utype += 5000000
                    if len(router.interfaces) == 1:
                        iasn = router.interfaces[0].asn
                        # if iasn > 0 and not (self.norelpeer and iasn in self.norelpeer) and not self.bgp.rel(iasn, sasn):
                        #     asn = None
                        #     utype = 0
            if not asn and len(router.succ) == 1:
                if len(router.interfaces) == 1:
                    isucc = peek(router.succ)  # single subsequent interface
                    sasn = self.iupdates.asn(isucc)  # annotation for subsequent interface
                    if sasn == -1:
                        sasn = isucc.asn
                    rasn = router.interfaces[0].asn
                    reltype = self.bgp.reltype(rasn, sasn)
                    if debug.DEBUG: print('One interface: {} -- {} == {}'.format(rasn, sasn, reltype))
                    if reltype != 1 and reltype != 2:
                        if sasn in router.dests and rasn not in router.dests:
                            asn = sasn
                            utype += 16000
            if not asn:
                for xasn in asns:
                    if xasn in router.dests and all(oasn in self.bgp.cone[xasn] for oasn in asns if xasn != oasn):
                        asn = xasn
                        utype += 36000
                        break
            # Tiebreaker 2 -- use only when tiebreaker 1 does not select an AS (most of the time)
            if not asn:
                if debug.DEBUG: print('Conesizes: {}'.format({a: self.bgp.conesize[a] for a in asns}))
                # First select from ASes that are both router origin ASes and subsequent ASes
                # Then select likely customer, based on smalles customer cone size
                asn = min(asns, key=lambda x: (not (x in sasn_origins[x] and x in sorigins), x not in router.dests, self.bgp.conesize[x], -x))
                utype += VOTE_TIE

        if asn not in iasns:
            overlap = {isucc.asn for isucc, isuccasns in router.origins.items() if isucc.asn in isuccasns and isucc.asn > 0}
            # overlap = iasns.keys() & sorigins.keys()
            if debug.DEBUG: print('Overlap: {}'.format(overlap))
            if len(overlap) == 1:
                if debug.DEBUG: print('Succs votes: {} < (2 * {}) / 3 = {}'.format(sorigins[asn], sum(sorigins.values()), (2 * sum(sorigins.values())) / 3))
                if succs[asn] < (2 * sum(sorigins.values())) / 3:
                    oasns = max_num(overlap, key=votes.__getitem__)
                    if len(oasns) == 1:
                        oasn = oasns[0]
                        if debug.DEBUG: print('Orgs: {} != {}'.format(self.as2org[oasn], self.as2org[asn]))
                        if self.as2org[oasn] != self.as2org[asn]:
                            asn = oasn
                            utype += 1000000

        # Check destination ASes
        # if not iasns and not router.nexthop and sum(succs.values()) <= 1:
        #     dasn, _ = self.annotate_lasthop(router)
        #     if debug.DEBUG:
        #         print('Curr ASN: {}'.format(asn))
        #         print('Last hop: {}'.format(dasn))
        #     if dasn != asn:
        #         if debug.DEBUG: print('{} in cone[{}]'.format(asn, dasn, asn in self.bgp.cone[dasn]))
        #         if asn in self.bgp.cone[dasn]:
        #             asn = dasn
        #             utype = 17000

        # Check for hidden AS
        # If no relationship between selected AS and an IR origin AS
        if iasns and all(asn != iasn and not self.bgp.rel(iasn, asn) for iasn in iasns):
            # for iasn in iasns:
            #     if iasn > 0 and iasn in router.dests:
            #         return iasn, 43
            # if not router.dests & votes.keys():
            #     dasns = {dasn for dasn in router.dests if any(iasn == dasn or self.bgp.rel(iasn, dasn) for iasn in iasns)}
            #     if debug.DEBUG:
            #         print('DASNs: {}'.format(dasns))
            #     if len(dasns) == 1:
            #         return peek(dasns), 42
            return self.hidden_asn(iasns, asn, utype, votes)
        return asn, utype

    def annotate_routers(self, routers: Collection[Router], usehints=False, use_provider=False, first=False, remap=False, increment=100000):
        pb = Progress(len(routers), 'Annotating routers', increment=increment)
        for router in pb.iterator(routers):
            asn = -1
            utype = -1
            if usehints and router.hints:
                asn, utype = self.annotate_router_hint(router, use_provider=use_provider)
            if asn <= 0:
                asn, utype = self.annotate_router(router, first=first, remap=False)
            self.rupdates.add_update(router, asn, self.as2org[asn], utype)

    def annotate_vrf_routers(self, routers: Collection[Router], increment=100000):
        pb = Progress(len(routers), 'Annotating forwarding routers', increment=increment)
        for router in pb.iterator(routers):
            asn, utype = self.annotate_router_vrf(router)
            self.rupdates.add_update_direct(router, asn, self.as2org[asn], utype)

    def annotate_interface(self, interface: Interface):
        edges: Dict[Router, int] = interface.pred
        if debug.DEBUG:
            print('ASN: {}'.format(interface.asn))
            print('VRF: {}'.format(interface.vrf))
        votes = Counter()
        for rpred, num in edges.items():
            asn = self.rupdates.asn(rpred)
            if debug.DEBUG: print('Router={}, RASN={}, ASN={}'.format(rpred.name, asn, rpred.interfaces[0].asn))
            votes[asn] += num
        if debug.DEBUG: print('Votes: {}'.format(votes))
        if len(votes) == 1:
            asn = peek(votes)
            utype = 1 if len(edges) > 1 else 0
        else:
            asns = max_num(votes, key=votes.__getitem__)
            if debug.DEBUG: print('MaxNum: {}'.format(asns))
            rels = [asn for asn in asns if interface.asn == asn or self.bgp.rel(interface.asn, asn)]
            if not rels:
                rels = asns
            if debug.DEBUG:
                print('Rels: {}'.format(rels))
                print('Sorted Rels: {}'.format(sorted(rels, key=lambda x: (
                    x != interface.asn, -self.bgp.provider_rel(interface.asn, x), -self.bgp.conesize[x], x))))
            # asn = max(asns, key=lambda x: (x == interface.asn, bdrmapit.bgp.conesize[x], -x))
            # asn = min(rels, key=lambda x: (x != interface.asn, -self.bgp.provider_rel(interface.asn, x), -self.bgp.conesize[x], x))
            asn = min(rels, key=lambda x: (x != interface.asn, -self.bgp.conesize[x], x))
            utype = 1 if len(asns) == 1 and len(edges) > 1 else 2
        if asn == -1:
            return -2, 2
        return asn, utype

    def annotate_interfaces(self, interfaces: Collection[Interface]):
        pb = Progress(len(interfaces), 'Adding links', increment=100000)
        for interface in pb.iterator(interfaces):
            if interface.asn >= 0:
                asn, utype = self.annotate_interface(interface)
                self.iupdates.add_update(interface, asn, self.as2org[asn], utype)

    def graph_refinement(self, routers: List[Router], interfaces: List[Interface], iterations=-1, vrfrouters: List[Router] = None, usehints=False, use_provider=False):
        self.previous_updates = []
        iteration = 0
        while iterations < 0 or iteration < iterations:
            Progress.message('********** Iteration {:,d} **********'.format(iteration), file=sys.stderr)
            self.annotate_routers(routers, first=(iteration == 0), usehints=usehints, use_provider=use_provider)
            self.rupdates.advance()
            if vrfrouters:
                self.annotate_vrf_routers(vrfrouters)
            self.annotate_interfaces(interfaces)
            self.iupdates.advance()
            ru = dict(self.rupdates)
            iu = dict(self.iupdates)
            if (ru, iu) in self.previous_updates:
                break
            self.previous_updates.append((ru, iu))
            iteration += 1



from collections import Counter
from itertools import chain
from typing import Optional

from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.utils.utils import peek

from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.node import Router
from bdrmapit.algorithm import debug


class RegexMixin:

    bgp: Optional[BGP] = None
    as2org: Optional[AS2Org] = None
    rupdates: Optional[Updates] = None

    def hidden_provider_hint(self, router):
        sasns = {succ.asn for succ in router.succ if succ.asn > 0}
        providers = {pasn for sasn in chain(sasns, router.dests) for pasn in self.bgp.providers[sasn]}
        provinter = providers & router.hints
        if len(provinter) == 1:
            utype = 0xff08
            return peek(provinter), utype
        return -1, -1

    # def hidden_provider_hint(self, router):
    #     iasns = {interface.asn for interface in router.interfaces if interface.asn > 0}
    #     if not iasns:
    #         iasns = set()
    #         for interface in router.interfaces:
    #             for pred in interface.pred:
    #                 update = self.rupdates[pred]
    #                 if update is not None:
    #                     iasn = update.asn
    #                     if iasn > 0:
    #                         iasns.add(iasn)
    #     sasns = {succ.asn for succ in router.succ if succ.asn > 0}
    #     provsucc = {provider for asn in sasns for provider in self.bgp.providers[asn]}
    #     customers = {customer for asn in iasns for customer in self.bgp.customers[asn]}
    #     provcust = provsucc & customers
    #     provinter = provcust & router.hints
    #     if not provinter:
    #         provdest = {provider for asn in router.dests for provider in self.bgp.providers[asn]}
    #         provcust = provdest & customers
    #         provinter = provcust & router.hints
    #     if len(provinter) == 1:
    #         utype = 0xff08
    #         return peek(provinter), utype
    #     return -1, -1

    def annotate_router_hint(self, router: Router, use_provider=False):
        utype = 0
        if debug.DEBUG: print('Hints: {}'.format(router.hints))
        # iasns = {interface.asn for interface in router.interfaces if interface.asn > 0}
        sasns = Counter(succ.asn for succ in router.succ if succ.asn > 0)
        possible = sasns.keys() | router.dests
        if debug.DEBUG:
            # print('IASNs: {}'.format(iasns))
            print('SASNs: {}'.format(sasns))
            print('Dests: {}'.format(router.dests))
            print('{} in possible: {}'.format(router.hints, bool(router.hints & possible)))
        if not possible and len(router.hints) == 1:
            utype = 0xff00
            return peek(router.hints), utype
        intersection = possible & router.hints
        if len(intersection) == 1:
            # print(utype)
            # if router.hints & iasns:
            #     utype |= 0xff01
            if router.hints & sasns.keys():
                utype |= 0xff02
            if router.hints & router.dests:
                utype |= 0xff04
            # print(utype)
            return peek(intersection), utype
        elif len(intersection) > 2:
            return -1, utype
        posorgs = {self.as2org[asn] for asn in possible}
        hintorgs = {self.as2org[asn] for asn in router.hints}
        interorgs = posorgs & hintorgs
        if debug.DEBUG:
            print('{} in possible orgs: {}'.format(hintorgs, bool(hintorgs & posorgs)))
        if interorgs:
            # if hintorgs & {self.as2org[asn] for asn in iasns}:
            #     utype |= 0xfe01
            if hintorgs & {self.as2org[asn] for asn in sasns}:
                utype |= 0xfe02
            if hintorgs & {self.as2org[asn] for asn in router.dests}:
                utype |= 0xfe04
            return peek(router.hints), utype
        if use_provider:
            asn, utype = self.hidden_provider_hint(router)
            if asn > 0:
                return asn, utype
        return 0, utype



#!/usr/bin/env python
import json
import os
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from enum import Enum

from jsonschema import validate
from traceutils2.as2org import AS2Org
from traceutils2.bgp import BGP
from traceutils2.radix.ip2as import create_table

from bdrmapit.algorithm.algorithm import Bdrmapit
from bdrmapit.container.container import Container
from bdrmapit.output.saveres import Save, ITDK
import bdrmapit.scripts.traceparser as tp

class ExecTypes(Enum):
    traceparser = 1
    bdrmapit_all = 2
    bdrmapit_graph = 3
    bdrmapit_config = 4

def set_bdrmapit_parser(parser: ArgumentParser):
    group = parser.add_argument_group('CAIDA AS2Org')
    group.add_argument('-b', '--as2org', required=True, help='CAIDA AS2Org filename')
    group.add_argument('-B', '--as2org-extra', help='Filename with additional siblings')
    group = parser.add_argument_group('CAIDA AS Relationships')
    group.add_argument('-r', '--rels', required=True, help='The CAIDA relationship file that indicates provider or peer relationships.')
    group.add_argument('-c', '--cone', required=True, help='The CAIDA customer cone file.')
    group.add_argument('-P', '--peeringdb', help='PeeringDB json file (recommended).')
    parser.add_argument('-R', '--routers', help='Alias resolution file in CAIDA ITDK format.')
    parser.add_argument('-I', '--max-iterations', default=5, type=int, help='Maximum number of iterations to run the graph refinement loop.')
    parser.add_argument('-H', '--as-hints', help='AS hints file.')
    parser.add_argument('--no-echos', action='store_true', help='Ignore echo-only addresses.')
    set_bdrmapit_parser_output(parser)

def set_bdrmapit_parser_output(parser: ArgumentParser):
    group = parser.add_argument_group('Output')
    group.add_argument('-s', '--sqlite', help='Output filename for sqlite3 output.')
    group.add_argument('-k', '--itdk', help='Output in ITDK nodes.as format.')

def run_from_config(args):
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'schema.json')) as f:
        schema = json.load(f)
    with open(args.config) as f:
        config = json.load(f)
    if not args.graph_only:
        schema['required'].extend(['as2org', 'as-rels'])
    validate(config, schema)
    args.ip2as = config['ip2as']
    if 'graph' in config:
        args.etype = ExecTypes.bdrmapit_graph
        args.graph = config['graph']
    else:
        args.wfiles = config['warts'].get('files') if 'warts' in config else None
        args.wfilelist = config['warts'].get('files-list') if 'warts' in config else None
        args.afiles = config['atlas'].get('files') if 'atlas' in config else None
        args.afilelist = config['atlas'].get('files-list') if 'atlas' in config else None
        args.jfiles = config['jsonwarts'].get('files') if 'jsonwarts' in config else None
        args.jfilelist = config['jsonwarts'].get('files-list') if 'jsonwarts' in config else None
        args.processes = config['processes']
        if args.graph_only:
            args.output = args.graph_only
            args.etype = ExecTypes.traceparser
        else:
            args.etype = ExecTypes.bdrmapit_all
    if args.etype != ExecTypes.traceparser:
        args.as2org = config['as2org']['as2org']
        args.as2org_extra = config['as2org'].get('additional')
        args.rels = config['as-rels']['rels']
        args.cone = config['as-rels']['cone']
        args.max_iterations = config['max_iterations']
        args.routers = config.get('aliases')
        args.as_hints = config.get('hints')
        args.peeringdb = config.get('peeringdb')

def main(args=None):
    if args is None:
        parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
        subs = parser.add_subparsers()

        tparser = subs.add_parser('parser')
        tp.set_parser(tparser, output=True)
        tparser.set_defaults(etype=ExecTypes.traceparser)

        bparser = subs.add_parser('all')
        tp.set_parser(bparser, group='Traceroute Parsing', output=False)
        set_bdrmapit_parser(bparser)
        bparser.set_defaults(etype=ExecTypes.bdrmapit_all)

        gparser = subs.add_parser('graph')
        gparser.add_argument('-g', '--graph', required=True, help='Pickle file with graph.')
        gparser.add_argument('-i', '--ip2as', required=True, help='Filename of prefix-to-AS mappings in CAIDA prefix2as format.')
        set_bdrmapit_parser(gparser)
        gparser.set_defaults(etype=ExecTypes.bdrmapit_graph)

        cparser = subs.add_parser('json')
        cparser.add_argument('-c', '--config', required=True, help='JSON config file.')
        cparser.add_argument('--graph-only', help='Only create the graph, then save it to the specified file.')
        set_bdrmapit_parser_output(cparser)
        cparser.set_defaults(etype=ExecTypes.bdrmapit_config)

        # parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=bmpack.__version__))

        args = parser.parse_args()

    if args.etype != ExecTypes.traceparser:
        if not (args.sqlite or args.itdk):
            print('Must specify output filename', file=sys.stderr)
            sys.exit(1)

    if args.etype == ExecTypes.bdrmapit_config:
        run_from_config(args)
        return main(args=args)

    if args.etype == ExecTypes.traceparser:
        tp.main(args=args)
        return

    ip2as = create_table(args.ip2as)
    as2org = AS2Org(args.as2org, additional=args.as2org_extra)
    if args.etype == ExecTypes.bdrmapit_all:
        args.output = None
        parseres = tp.main(args=args, ip2as=ip2as)
        prep = Container(ip2as, as2org, parseres)
    else:
        sys.stdout.write('Unpickling graph.')
        prep = Container.load(ip2as, as2org, args.graph)
        sys.stdout.write(' Done.\n')

    bgp = BGP(args.rels, args.cone)
    use_hints = args.as_hints is not None

    graph = prep.construct(nodes_file=args.routers, hints_file=args.as_hints, no_echos=args.no_echos)

    bdrmapit = Bdrmapit(graph, as2org, bgp, strict=False)
    if args.peeringdb:
        bdrmapit.peeringdb_ixpasns(args.peeringdb, ip2as)
    bdrmapit.set_dests()
    bdrmapit.annotate_lasthops(usehints=use_hints, use_provider=True)
    bdrmapit.graph_refinement(bdrmapit.routers_succ, bdrmapit.interfaces_pred, iterations=args.max_iterations, usehints=use_hints, use_provider=True)

    if args.sqlite:
        save = Save(args.sqlite, bdrmapit, replace=True)
        save.save_annotations()
        save.save_ixps()
        save.save_links()
    if args.itdk:
        include_all = args.routers is None
        save = ITDK(bdrmapit)
        save.write_nodes(args.itdk, include_all=include_all)

if __name__ == '__main__':
    main()



from traceutils2.utils.dicts cimport EmptyDict, ZeroDict

cpdef enum RelType:
    provider = 1
    customer = 2
    peer = 3
    none = 4

cdef class BGP:
    cdef readonly set rels
    cdef readonly EmptyDict customers
    cdef readonly EmptyDict providers
    cdef readonly EmptyDict peers
    cdef readonly EmptyDict cone
    cdef readonly ZeroDict conesize

    cpdef bint customer_rel(self, int x, int y) except *;
    cpdef set multi_customers(self, asns);
    cpdef set multi_peers(self, asns);
    cpdef set multi_providers(self, asns);
    cpdef bint multi_rels(self, asn, others) except *;
    cpdef bint peer_rel(self, int x, int y) except *;
    cpdef bint provider_rel(self, int x, int y) except *;
    cpdef bint rel(self, int x, int y) except *;
    cpdef RelType reltype(self, int x, int y);



#!/usr/bin/env python
import pickle
from argparse import ArgumentParser
from enum import Enum
from multiprocessing.pool import Pool
from typing import Optional, List, Dict

from file2 import fopen
from pb_amarder.bar import Progress
from traceutils2.radix.ip2as import IP2AS, create_table, create_private
from traceutils2.scamper.atlas import AtlasReader
from traceutils2.scamper.hop import ICMPType, Hop
from traceutils2.scamper.warts import WartsReader, WartsJsonReader
from traceutils2.scamper.pyatlas import AtlasReader as AtlasOddReader

print('current')

_ip2as: Optional[IP2AS] = None
_filemap4: Optional[Dict[str, str]] = None
_filemap6: Optional[Dict[str, str]] = None
_prune_private = True

class OutputType(Enum):
    WARTS = 1
    ATLAS = 2
    ATLAS_ODD = 3
    JSONWARTS = 4

class TraceFile:
    def __init__(self, filename, type):
        self.filename = filename
        self.type = type

    def __repr__(self):
        return self.filename

def parse(tfile: TraceFile):
    results = set()
    if tfile.type == OutputType.WARTS:
        f = WartsReader(tfile.filename, ping=False)
    elif tfile.type == OutputType.ATLAS:
        f = AtlasReader(tfile.filename)
    elif tfile.type == OutputType.ATLAS_ODD:
        f = AtlasOddReader(tfile.filename)
    elif tfile.type == OutputType.JSONWARTS:
        f = WartsJsonReader(tfile.filename)
    else:
        raise Exception('Invalid output type: {}.'.format(tfile.type))
    try:
        f.open()
        for trace in f:
            if _prune_private:
                trace.prune_private(_ip2as)
            trace.prune_dups()
            trace.prune_loops(True)
            # if trace.loop:
            #     results.cycles.update(trace.loop)
            hops: List[Hop] = [h for h in trace.hops if _ip2as[h.addr] != -1 and h.addr != trace.src]
            if not hops: continue
            fhop: Hop = hops[0]
            lhop: Hop = hops[-1]
            for i in range(len(hops)):
                x: Hop = hops[i]
                if i == len(hops) - 1:
                    break
                y: Hop = hops[i+1]
                # if y.type == ICMPType.echo_reply or y.type == ICMPType.portping:
                #     break
                if y.type == ICMPType.spoofing and y.icmp_q_ttl > 1:
                    break
                distance = y.probe_ttl - x.probe_ttl
                if y.icmp_q_ttl == 0:
                    distance += 1
                if distance == 1:
                    results.add((x.addr, y.addr, y.icmp_type))
    except UnicodeDecodeError:
        print(tfile.filename)
        raise
    except EOFError:
        print(tfile.filename)
    finally:
        f.close()
    return results

def parse_sequential(files):
    results = set()
    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: '{:,d}'.format(len(results)))
    for tfile in pb.iterator(files):
        newresults = parse(tfile)
        results.update(newresults)
    return results

def parse_parallel(files, poolsize):
    results = set()
    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: '{:,d}'.format(len(results)))
    with Pool(poolsize) as pool:
        for newresults in pb.iterator(pool.imap_unordered(parse, files)):
            results.update(newresults)
    return results

def run(files, ip2as: IP2AS, poolsize, output=None, prune_private=True, serialize=False):
    global _ip2as, _filemap4, _filemap6, _prune_private
    _ip2as = ip2as
    _prune_private = prune_private

    poolsize = min(len(files), poolsize)
    print(poolsize)
    results = parse_parallel(files, poolsize) if poolsize != 1 else parse_sequential(files)
    te = set()
    echo = set()
    for x, y, itype in results:
        if itype == 0:
            echo.add((x, y))
        else:
            te.add((x, y))
    echo -= te
    adjs = {(x, y, True) for x, y in te} | {(x, y, False) for x, y in echo}
    if output:
        if serialize:
            with open(output, 'wb') as f:
                pickle.dump(adjs, f)
        else:
            with fopen(output, 'wt') as f:
                for x, y, z in adjs:
                    f.write('{}\t{}\t{}\n'.format(x, y, int(z)))
    return results

def read_filemap(filename):
    filemap = {}
    with open(filename) as f:
        for line in f:
            if line[0] == '#':
                continue
            file, addr = line.split()
            filemap[file] = addr
    return filemap

def main():
    parser = ArgumentParser()
    parser.add_argument('-w', '--wfiles', help='File with list of newline-separated filenames.')
    parser.add_argument('-W', '--wfilelist', nargs='+', help='List of filenames, space separated.')
    parser.add_argument('-a', '--afiles', help='File with list of newline-separated filenames.')
    parser.add_argument('-A', '--afilelist', nargs='+', help='List of filenames, space separated.')
    parser.add_argument('-j', '--jfiles')
    parser.add_argument('-J', '--jfilelist', nargs='+')
    parser.add_argument('-p', '--poolsize', type=int, default=1)
    parser.add_argument('-o', '--output', required=True)
    parser.add_argument('-k', '--keep-private', action='store_true')
    parser.add_argument('-P', '--pickle', action='store_true')
    args = parser.parse_args()
    files = []
    if args.wfiles:
        with fopen(args.wfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.WARTS) for line in f if line[0] != '#')
    if args.wfilelist:
        files.extend(TraceFile(file, OutputType.WARTS) for file in args.wfilelist)
    if args.afiles:
        with fopen(args.afiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.ATLAS) for line in f if line[0] != '#')
    if args.afilelist:
        files.extend(TraceFile(file, OutputType.ATLAS) for file in args.afilelist)
    if args.jfiles:
        with fopen(args.jfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.JSONWARTS) for line in f if line[0] != '#')
    if args.jfilelist:
        files.extend(TraceFile(file, OutputType.JSONWARTS) for file in args.jfilelist)
    ip2as = create_private()
    prune_private = not args.keep_private
    run(files, ip2as, args.poolsize, args.output, prune_private=prune_private, serialize=args.pickle)

if __name__ == '__main__':
    main()



from traceutils2.radix.radix_prefix cimport RadixPrefix


cdef class RadixNode:
    def __init__(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None):
        if prefix is not None:
            self.prefix = prefix
            self.bitlen = self.prefix.bitlen
        else:
            self.prefix = None
            self.bitlen = prefix_size
            self.parent = parent
        self.left = None
        self.right = None

    def __str__(self):
        return str(self.prefix)

    def __repr__(self):
        return 'Prefix<{}>'.format(self.prefix)

    @property
    def network(self):
        return self.prefix.network()

    @property
    def prefixlen(self):
        return self.bitlen

    @property
    def family(self):
        return self.prefix.family

    @property
    def packed(self):
        return self.prefix.packed

cdef class RadixNodeASN(RadixNode):
    def __init__(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None, int asn=0):
        super().__init__(prefix, prefix_size, parent)
        self.asn = asn

    def __repr__(self):
        return '<{} ASN={}>'.format(self.prefix, self.asn)

cdef class RadixNodeASNs(RadixNode):
    def __init__(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None, list asns=None):
        super().__init__(prefix, prefix_size, parent)
        self.asns = asns

    def __repr__(self):
        return '<{} ASNs={}>'.format(self.prefix, self.asns)

cdef class RadixNodeData(RadixNode):
    def __init__(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None, **kwargs):
        super().__init__(prefix, prefix_size, parent)
        self.data = kwargs

    def __repr__(self):
        return '<{} Data={}>'.format(self.prefix, self.data)



from typing import Optional

from traceutils2.bgp.bgp import BGP

from bdrmapit.algorithm.updates_dict import Updates


class HelpersMixin:

    rupdates: Optional[Updates] = None
    bgp: Optional[BGP] = None

    def multi_customers(self, asns):
        return {customer for asn in asns for customer in self.bgp.customers[asn]}

    def multi_peers(self, asns):
        return {peer for asn in asns for peer in self.bgp.peers[asn]}

    def multi_providers(self, asns):
        return {provider for asn in asns for provider in self.bgp.providers[asn]}

    def any_rels(self, asn, others):
        for other in others:
            if self.bgp.rel(asn, other):
                return True
        return False



from traceutils2.scamper.hop cimport Hop, Trace, Reader

cdef class AtlasHop(Hop):
    cdef public dict icmpext
    cdef public str flags
    cdef public int late
    cdef public int dup
    cdef public str edst
    cdef list hdropts

cdef class AtlasTrace(Trace):
    cdef public int af, fw, group_id, lts, msm_id, paris_id, prb_id, size
    cdef public str dst_addr, dst_name, msm_name, proto, src_addr, type
    cdef public long endtime, timestamp
    cdef public list result

cdef class AtlasReader(Reader):
    # cdef str filename
    cdef object f



from traceutils2.as2org.as2org cimport AS2Org
from traceutils2.radix.ip2as cimport IP2AS

cdef class Graph:
    cdef readonly dict interfaces, routers

cpdef Graph construct_graph(list addrs, dict nexthop, dict multi, dict dps, list mpls, IP2AS ip2as, AS2Org as2org, str nodes_file=*, int increment=*);



from collections import Counter, defaultdict
from typing import Optional, List, Set

from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.utils.utils import max_num

from bdrmapit.algorithm import debug
from bdrmapit.algorithm.utypes import VOTE_SINGLE, VOTE_TIE
from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.node import Router
from bdrmapit.vrf.vrfedge import VRFEdge, VType


class VRFMixin:

    rupdates: Optional[Updates] = None
    bgp: Optional[BGP] = None
    as2org: Optional[AS2Org] = None

    def vrf_heuristics(self, edge: VRFEdge, origins: Set[int]):
        rsucc: Router = edge.node
        vtype = edge.vtype
        if debug.DEBUG: print('VType={}'.format(vtype.name))
        for iface in rsucc.interfaces:
            if iface.asn > 0 and iface.asn in origins:
                return iface.asn
        rsucc_asn = self.rupdates.asn(rsucc)
        return rsucc_asn

    def annotate_router_vrf(self, router: Router):
        edge: VRFEdge
        utype: int = 0

        iasns = Counter(interface.asn for interface in router.interfaces if interface.asn > 0)
        if debug.DEBUG:
            print('IASN: {}'.format(iasns))
            print('Edges={}, NH={}'.format(len(router.succ), router.nexthop))
            print('VRF={}'.format(router.vrf))

        # Use heuristics to determine link votes
        vtype = None
        succs = Counter()
        nonvrf = Counter()
        sasn_origins = defaultdict(set)
        for edge in router.succ:
            origins = router.origins[edge]
            if debug.DEBUG: print('Succ={}, ASN={}, VRF={}'.format(edge.node.name, self.rupdates[edge.node], edge.node.vrf))
            # succ_asn = self.vrf_heuristics(edge, origins, iasns)
            succ_asn = self.vrf_heuristics(edge, origins)
            if vtype is None:
                vtype = edge.vtype
            elif vtype.value != edge.vtype.value:
                vtype = VType.both
            if debug.DEBUG: print('Heuristic: {}'.format(succ_asn))
            if succ_asn > 0:
                succs[succ_asn] += 1
                sasn_origins[succ_asn].update(origins)
                if not edge.node.vrf:
                    nonvrf[succ_asn] += 1
        if debug.DEBUG:
            print('Succs: {}'.format(succs))
            print('VType: {}'.format(vtype))

        # Create votes counter and add interface AS
        if debug.DEBUG: print('NonVRF: {}'.format(nonvrf))
        if nonvrf:
            votes = nonvrf
            utype += 50000
        else:
            votes = succs + iasns
        if debug.DEBUG: print('Votes: {}'.format(votes))
        if not votes:
            return -1, -1

        votes_rels: List[int] = [vasn for vasn in votes if any(
            vasn == iasn or self.bgp.rel(iasn, vasn) or self.as2org[iasn] == self.as2org[vasn] for iasn in iasns)]
        if debug.DEBUG: print('Vote Rels: {}'.format(votes_rels))
        if len(votes_rels) < 2:
            asns = max_num(votes, key=votes.__getitem__)
            if debug.DEBUG: print('ASNs: {}'.format(asns))
        else:
            for vasn in list(votes):
                if vasn not in votes_rels:
                    for vr in votes_rels:
                        if self.as2org[vr] == self.as2org[vasn]:
                            votes[vr] += votes.pop(vasn, 0)
            asns = max_num(votes_rels, key=votes.__getitem__)
            othermax = max(votes, key=votes.__getitem__)
            if debug.DEBUG:
                print('ASNs: {}'.format(asns))
                print('Othermax: {}'.format(othermax))
            if router.nexthop and votes[othermax] > votes[asns[0]] * 4:
                utype += 3000
                return othermax, utype
        if len(asns) == 1:
            asn = asns[0]
            utype += VOTE_SINGLE
        else:
            asn = min(asns, key=lambda x: (self.bgp.conesize[x], -x))
            utype += VOTE_TIE
        return asn, utype

    def sort_vrf(self, router: Router):
        nedges = len(router.succ)
        iasns = {interface.asn for interface in router.interfaces}
        iasn = min(iasns, key=lambda x: (self.bgp.conesize[x], -x))
        conesize = self.bgp.conesize[iasn]
        return -nedges, conesize, -iasn



cdef class Router:
    cdef:
        readonly str name
        readonly list interfaces
        public bint nexthop
        public bint vrf
        readonly set succ
        readonly set dests
        readonly dict origins
        public bint echo
        public bint cycle
        public set hints

    cpdef Router copy(self);

cdef class Interface:
    cdef:
        readonly str addr
        readonly int asn
        readonly str org
        public Router router
        readonly dict pred
        public set dests
        public bint vrf
        public bint echo
        public bint cycle
        public int hint

    cpdef Interface copy(self);

ctypedef fused Node:
    Router
    Interface


from collections import defaultdict

from file2 import fopen

from traceutils2.utils.dicts cimport EmptyDict, ZeroDict

cdef class BGP:

    def __init__(self, str rels, str cone, str extras = None):
        cdef str line
        cdef int x, y, rel
        cdef list customers
        self.rels = set()
        _customers = defaultdict(set)
        _providers = defaultdict(set)
        _peers = defaultdict(set)
        self.cone = EmptyDict()
        self.conesize = ZeroDict()
        with fopen(rels) as f:
            for line in f:
                if not line.startswith('#'):
                    x, y, rel = map(int, line.strip().split('|'))
                    self.rels.add((x, y))
                    self.rels.add((y, x))
                    if rel == -1:
                        _customers[x].add(y)
                        _providers[y].add(x)
                    elif rel == 0:
                        _peers[x].add(y)
                        _peers[y].add(x)
        if extras is not None:
            with fopen(extras) as f:
                for line in f:
                    if not line.startswith('#'):
                        x, y, rel = map(int, line.strip().split('|'))
                        self.rels.add((x, y))
                        self.rels.add((y, x))
                        if rel == -1:
                            _customers[x].add(y)
                            _providers[y].add(x)
                        elif rel == 0:
                            _peers[x].add(y)
                            _peers[y].add(x)
        with fopen(cone) as f:
            for line in f:
                if not line.startswith('#'):
                    provider, *customers = map(int, line.split())
                    self.cone[provider] = set(customers)
                    self.conesize[provider] = len(customers)
        self.customers = EmptyDict(_customers)
        self.providers = EmptyDict(_providers)
        self.peers = EmptyDict(_peers)

    cpdef bint customer_rel(self, int x, int y) except *:
        return x in self.customers[y]

    cpdef set multi_customers(self, asns):
        return {a for asn in asns for a in self.customers[a]}

    cpdef set multi_peers(self, asns):
        return {a for asn in asns for a in self.peers[a]}

    cpdef set multi_providers(self, asns):
        return {a for asn in asns for a in self.providers[a]}

    cpdef bint multi_rels(self, asn, others) except *:
        for other in others:
            if self.rel(asn, other):
                return True
        return False

    cpdef bint peer_rel(self, int x, int y) except *:
        return y in self.peers[x]

    cpdef bint provider_rel(self, int x, int y) except *:
        return x in self.providers[y]

    cpdef bint rel(self, int x, int y) except *:
        return (x, y) in self.rels

    cpdef RelType reltype(self, int x, int y):
        if y in self.customers[x]:
            return RelType.provider
        elif y in self.providers[x]:
            return RelType.customer
        elif y in self.peers[x]:
            return RelType.peer
        return RelType.none



from collections import Counter
from typing import Set, Optional, Collection

from pb_amarder import Progress
from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.utils.utils import peek

from bdrmapit.algorithm import debug
from bdrmapit.algorithm.regexmixin import RegexMixin
from bdrmapit.algorithm.utypes import NODEST, MISSING_NOINTER, HEAPED
from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Router


class LastHopsMixin(RegexMixin):

    rupdates: Optional[Updates] = None
    bgp: Optional[BGP] = None
    as2org: Optional[AS2Org] = None
    graph: Optional[Graph] = None
    strict = False
    lasthops: Optional[Collection[Router]] = None

    def set_dests(self, increment=1000000):
        """
        Set destination AS sets for each router, and remove potential relocated prefixes for last hop interfaces.
        :param increment: status increment
        """
        modified = 0
        pb = Progress(len(self.graph.routers), 'Setting destinations', increment=increment, callback=lambda: 'Modified {:,d}'.format(modified))
        for router in pb.iterator(self.graph.routers.values()):
            for interface in router.interfaces:
                # Copy destination ASes to avoid messing up original
                idests: Set[int] = set(interface.dests)
                # If last hop, interface has non-IXP AS mapping, and interface has destination ASes
                if not router.succ and idests and interface.asn > 0:
                    origin = interface.asn
                    # Interface must have exactly 2 destination ASes and one must be its origin AS
                    if len(idests) == 2 and origin in idests:
                        other_asn = peek(idests - {origin})  # other AS
                        # If other AS is likely customer of interface origin AS, and it's a small AS
                        if self.bgp.conesize[origin] > self.bgp.conesize[other_asn] and self.bgp.conesize[other_asn] < 5:
                            idests.discard(origin)
                            modified += 1
                # Add all remaining destination ASes to the router destination AS set
                router.dests.update(idests)

    def annotate_lasthop_nodests(self, iasns):
        if debug.DEBUG: print('No dests')
        # No interface origin ASes. Only cause is the addresses had no matching prefix.
        if len(iasns) == 0:
            # No AS mapping possible with current method.
            # TODO: Might be worth looking backward for mapping. Need to investigate how common.
            return -1, 1
        # Single router origin AS, so select it.
        elif len(iasns) == 1:
            return peek(iasns), 2
        # Collect ASes that have a relationship to all other ASes.
        allrels = {iasn for iasn in iasns if all(self.as2org[iasn] == self.as2org[oasn] or self.bgp.rel(iasn, oasn) for oasn in iasns if iasn != oasn)}
        if allrels:
            if debug.DEBUG:
                for asn in allrels:
                    print('{}: {}'.format(asn, iasns[asn]))
            # Select AS (with relationship to all others) with the most votes, then smallest customer cone.
            return max(allrels, key=lambda x: (iasns[x], -self.bgp.conesize[x], x)), 3
        # No AS is directly connected (according to BGP) with every other origin AS.
        # Try to select a single AS that is a customer of all origin ASes.
        hidden = []  # List of customer sets
        for iasn in iasns:
            # rels = self.bgp.providers[iasn] | self.bgp.peers[iasn] | self.bgp.customers[iasn]
            rels = self.bgp.customers[iasn]
            hidden.append(rels)
        # Take intersection of all customer sets
        intersection: Set[int] = hidden[0]
        for rels in hidden[1:]:
            intersection.intersection_update(rels)
        # If single AS intersection, select it
        if len(intersection) == 1:
            return peek(intersection), 4
        if debug.DEBUG:
            for iasn in iasns:
                print('{}: {}'.format(iasn, iasns[iasn]))
        # Select the most frequent origin AS, break ties with smallest customer cone size
        return max(iasns, key=lambda x: (iasns[x], -self.bgp.conesize[x], x)), 5

    def annotate_lasthop_norels(self, dests, iasns):
        if self.strict:
            return max(iasns, key=lambda x: (iasns[x], -self.bgp.conesize[x], x)), NODEST
        if iasns:
            if debug.DEBUG: print('IASNs: {}'.format(iasns))
            if debug.DEBUG:
                print('Providers: {}'.format(self.multi_providers(dests)))
            intersection = self.multi_providers(dests) & self.multi_customers(iasns)
            if len(intersection) == 1:
                if debug.DEBUG: print('Inter Cust: {}'.format(intersection))
                return peek(intersection), 10000
            # intersection = self.multi_providers(dests) & self.multi_peers(iasns)
            # if len(intersection) == 1:
            #     return peek(intersection), 30000
            intersection = self.multi_customers(dests) & self.multi_providers(iasns)
            if len(intersection) == 1:
                return peek(intersection), 20000
        # asn = min(dests, key=lambda x: (self.bgp.conesize[x], -x))
        asn = max(dests, key=lambda x: (self.bgp.conesize[x], -x))
        return asn, MISSING_NOINTER

    def annotate_lasthop(self, router: Router, dests=None):
        if dests is None:
            dests = router.dests
        iasns = Counter(interface.asn for interface in router.interfaces if interface.asn > 0)
        if debug.DEBUG:
            print('IASNs: {}'.format(iasns))
            print('Dests: {}'.format(dests))
        # No destination ASes
        if len(router.dests) == 0 or all(dest <= 0 for dest in router.dests):
            return self.annotate_lasthop_nodests(iasns)
        # Use overlapping ASes if available
        overlap = iasns.keys() & dests
        if debug.DEBUG: print('Dest IASN intersection: {}'.format(overlap))
        if overlap:
            if len(overlap) == 1:
                return peek(overlap), HEAPED
            return min(overlap, key=lambda x: (self.bgp.conesize[x], -x)), HEAPED
        # No overlapping ASes, use relationship ASes
        rels = {dasn for dasn in dests if self.any_rels(dasn, iasns)}
        if debug.DEBUG:
            print({(dasn, self.bgp.conesize[dasn]) for dasn in dests})
        if debug.DEBUG: print('Rels: {}'.format(rels))
        if rels:
            # Select overlapping or relationship AS with largest customer cone
            # return min(rels, key=lambda x: (self.bgp.conesize[x], -x)), HEAPED
            if len(rels) >= 4:
                return max(iasns, key=lambda x: sum(self.bgp.rel(x, dasn) for dasn in rels)), HEAPED
            maxasn = max(rels, key=lambda x: (self.bgp.conesize[x], -x))
            if len(dests - self.bgp.cone[maxasn]) > 4:
                if debug.DEBUG:
                    print('Uncovered dests for {}: {}'.format(maxasn, dests - self.bgp.cone[maxasn]))
                return max(iasns, key=lambda x: sum(self.bgp.rel(x, dasn) for dasn in rels)), HEAPED
            return maxasn, HEAPED
            # return max(rels, key=lambda x: (len(self.bgp.cone[x] & dests), -x)), HEAPED
        # No relationship between any origin AS and any destination AS
        return self.annotate_lasthop_norels(dests, iasns)

    def annotate_lasthops(self, routers=None, usehints=False, use_provider=False):
        if routers is None:
            routers = self.lasthops
        pb = Progress(len(routers), message='Last Hops', increment=100000)
        for router in pb.iterator(routers):
            asn = -1
            utype = -1
            if usehints and router.hints:
                asn, utype = self.annotate_router_hint(router, use_provider=use_provider)
            if asn <= 0:
                asn, utype = self.annotate_lasthop(router)
            self.rupdates.add_update_direct(router, asn, self.as2org[asn], utype)



import pickle
from typing import Dict, Union

from file2 import fopen
from pb_amarder import Progress

from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Interface, Router
from bdrmapit.vrf.vrfedge import VRFEdge


class Container:
    def __init__(self, ip2as, as2org, addrs=None, nexthop=None, multi=None, dps=None, mpls=None, spoofing=None, echos=None, cycles=None):
        self.ip2as = ip2as
        self.as2org = as2org
        self.addrs = addrs
        self.nexthop = nexthop
        self.multi = multi
        self.dps = dps
        self.mpls = mpls
        self.spoofing = spoofing
        self.echos = echos
        self.cycles = cycles
        self.interfaces: Dict[str, Interface] = {}
        self.routers: Dict[str, Router] = {}

    @classmethod
    def load(cls, filename, ip2as, as2org):
        with open(filename, 'rb') as f:
            results = pickle.load(f)
        return cls(ip2as, as2org, **results)

    def alladdrs(self):
        return set(self.addrs) | set(self.echos)

    def create_node(self, addr, router: Router):
        """
        Create new interface node and assign it to router.
        :param addr: address of new interface node
        :param router: router node representing the interface's router
        """
        asn = self.ip2as.asn(addr)
        # Make sure address is not from private address space
        if asn >= 0 or asn <= -100:
            # Create interface
            interface = Interface(addr, asn, self.as2org[asn])
            self.interfaces[addr] = interface
            interface.router = router
            # Add interface to router
            router.interfaces.append(interface)
            self.routers[router.name] = router
            # interface.echo = echo
            # interface.cycle = cycle

    def create_nodes(self, nodes_file, increment=100000):
        """
        Create router nodes based on alias resolution.
        :param nodes_file: filename containing alias resolution groupings in CAIDA format
        :param increment: increment for status
        """
        taddrs = set(self.addrs)
        pb = Progress(message='Creating nodes', increment=increment, callback=lambda: 'Routers {:,d} Interfaces {:,d}'.format(len(self.routers), len(self.interfaces)))
        with fopen(nodes_file, 'rt') as f:
            for line in pb.iterator(f):
                if line[0] != '#':
                    _, nid, *naddrs = line.split()
                    if not any(addr in taddrs for addr in naddrs):
                        continue
                    nid = nid[:-1]
                    router = Router(nid)
                    self.routers[router.name] = router
                    for addr in naddrs:
                        self.create_node(addr, router)

    def create_remaining(self, aliases: bool, increment=100000):
        """
        Create router nodes for any interfaces not seen in the alias resolution dataset, or when there is not alias resolution dataset.
        :param addrs: addresses included in the graph
        :param aliases: flag to indicate if alias resoultion was used
        :param increment: increment for status
        """
        pb = Progress(len(self.addrs), 'Creating remaining routers and interfaces', increment=increment)
        for addr in pb.iterator(self.addrs):
            if not aliases or addr not in self.interfaces:
                router = Router(addr)
                self.create_node(addr, router)

    @staticmethod
    def add_succ(router: Router, interface: Interface, succ: Union[VRFEdge, Interface]):
        if succ in router.succ:
            origins = router.origins[succ]
            origins.add(interface.asn)
        else:
            router.succ.add(succ)
            router.origins[succ] = {interface.asn}

    @staticmethod
    def add_pred(interface: Interface, prouter: Router):
        predcount = interface.pred.get(prouter, 0)
        interface.pred[prouter] = predcount + 1

    def add_nexthop(self, increment=100000):
        """
        Add nexthop edges.
        :param nexthop: nexthop edges
        :param increment: increment for status
        """
        pb = Progress(len(self.nexthop), 'Adding nexthop edges', increment=increment)
        for addr, edges in pb.iterator(self.nexthop.items()):
            interface = self.interfaces[addr]
            router = interface.router
            router.nexthop = True
            for edge in edges:
                succ = self.interfaces[edge]
                if succ.router != router:
                    self.add_succ(router, interface, succ)
                    self.add_pred(succ, router)

    def add_multi(self, increment=100000):
        """
        Add multiple hop edges.
        :param multi: multiple hop edges
        :param increment: increment for status
        """
        pb = Progress(len(self.multi), 'Adding multihop edges', increment=increment)
        for addr in pb.iterator(self.multi):
            interface = self.interfaces[addr]
            router = interface.router
            if not router.nexthop:
                edges = self.multi[addr]
                for edge in edges:
                    succ = self.interfaces[edge]
                    if succ.router != router:
                        self.add_succ(router, interface, succ)

    def add_dests(self, increment=100000):
        """
        Add destination ASes for each interface and router.
        :param dps: interface to destination mappings
        :param increment: increment for status
        """
        pb = Progress(len(self.dps), 'Adding destination ASes', increment=increment)
        for addr, dests in pb.iterator(self.dps.items()):
            interface = self.interfaces.get(addr)
            if interface is not None:
                interface.dests.update(dests)

    def create_graph(self):
        """
        Create the graph based on the interfaces, routers, and edges.
        :return: the graph
        """
        return Graph(interfaces=self.interfaces, routers=self.routers)

    def reset(self):
        self.interfaces = {}
        self.routers = {}

    def add_echos(self, increment=1000000):
        echos = 0
        pb = Progress(len(self.echos), 'Adding and marking echos', increment=increment, callback=lambda: '{:,d}'.format(echos))
        for addr in pb.iterator(self.echos):
            if addr in self.interfaces:
                interface = self.interfaces[addr]
                if not interface.dests:
                    interface.echo = True
                    # if all(i.echo for i in interface.router.interfaces):
                    #     interface.router.echo = True
            else:
                router = Router(addr)
                # router.echo = True
                self.create_node(addr, router)
                self.interfaces[addr].echo = True

    def add_cycles(self, increment=1000000):
        echos = 0
        pb = Progress(len(self.cycles), 'Adding and marking cycles', increment=increment, callback=lambda: '{:,d}'.format(echos))
        for addr in pb.iterator(self.echos):
            if addr in self.interfaces:
                interface = self.interfaces[addr]
                if not interface.dests:
                    interface.cycle = True
                    if all(i.cycle for i in interface.router.interfaces):
                        interface.router.echo = True
            else:
                router = Router(addr)
                router.cycle = True
                self.create_node(addr, router, cycle=True)

    def construct(self, nodes_file=None, echos=False, cycles=False):
        """
        Construct the graph from scratch.
        :param addrs: addresses seen in the dataset
        :param nexthop: nexthop edges
        :param multi: multiple hop edges
        :param dps: interface to destination ASes
        :param nodes_file: alias resolution dataset
        :return: the graph
        """
        if nodes_file is not None:
            self.create_nodes(nodes_file=nodes_file)
        self.create_remaining(nodes_file is not None)
        self.add_nexthop()
        self.add_multi()
        self.add_dests()
        if echos:
            self.add_echos()
        if cycles:
            self.add_cycles()
        return self.create_graph()



#!/usr/bin/env python
from argparse import ArgumentParser
from enum import Enum
from multiprocessing.pool import Pool
from typing import Optional, List, Dict

from file2 import fopen
from pb_amarder import Progress
from traceutils2.radix.ip2as import IP2AS, create_private
from traceutils2.scamper.atlas import AtlasReader
from traceutils2.scamper.hop import ICMPType
from traceutils2.scamper.warts import WartsReader, WartsJsonReader
from traceutils2.scamper.pyatlas import AtlasReader as AtlasOddReader
from traceutils2.utils.net import otherside, prefix_addrs

print('test')

_ip2as: Optional[IP2AS] = None
_filemap4: Optional[Dict[str, str]] = None
_filemap6: Optional[Dict[str, str]] = None
_prune_loops = False
_noechos = False
_subnet = False
_include_dsts = False
_fours = False

class OutputType(Enum):
    WARTS = 1
    ATLAS = 2
    ATLAS_ODD = 3
    JSONWARTS = 4

class TraceFile:
    def __init__(self, filename, type):
        self.filename = filename
        self.type = type

    def __repr__(self):
        return self.filename

def parse(tfile: TraceFile):
    addrs = set()
    if tfile.type == OutputType.WARTS:
        f = WartsReader(tfile.filename, ping=False)
    elif tfile.type == OutputType.ATLAS:
        f = AtlasReader(tfile.filename)
    elif tfile.type == OutputType.ATLAS_ODD:
        f = AtlasOddReader(tfile.filename)
    elif tfile.type == OutputType.JSONWARTS:
        f = WartsJsonReader(tfile.filename)
    else:
        raise Exception('Invalid output type: {}.'.format(tfile.type))
    try:
        f.open()
        fiter = iter(f)
        while True:
            try:
                trace = next(fiter)
            except StopIteration:
                break
            except (OSError, UnicodeDecodeError, EOFError):
                print(tfile.filename)
                break
            if _include_dsts:
                addrs.add(trace.dst)
            trace.prune_private(_ip2as)
            trace.prune_dups()
            if _prune_loops:
                trace.prune_loops(True)
            if _fours:
                for hop in trace.hops:
                    if hop.type != ICMPType.echo_reply:
                        addr = hop.addr
                        addrs.update(prefix_addrs(addr, 2))
            else:
                addrs.update(hop.addr for hop in trace.hops if not (_noechos and hop.type == ICMPType.echo_reply))
                if _subnet:
                    for hop in trace.hops:
                        if not (_noechos and hop.type == ICMPType.echo_reply):
                            addr = hop.addr
                            try:
                                other4 = otherside(addr, 4)
                                addrs.add(other4)
                            except:
                                pass
                            other2 = otherside(addr, 2)
                            addrs.add(other2)
    finally:
        f.close()
    return addrs

def parse_sequential(files):
    results = set()
    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: '{:,d}'.format(len(results)))
    for tfile in pb.iterator(files):
        results.update(parse(tfile))
    return results

def parse_parallel(files, poolsize):
    results = set()
    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: '{:,d}'.format(len(results)))
    with Pool(poolsize) as pool:
        for newresults in pb.iterator(pool.imap_unordered(parse, files)):
            results.update(newresults)
    return results

def run(files, ip2as: IP2AS, poolsize, output=None, prune_loops=False, noechos=False, subnet=False, include_dsts=False, fours=False):
    global _ip2as, _prune_loops, _noechos, _subnet, _include_dsts, _fours
    _ip2as = ip2as
    _prune_loops = prune_loops
    _noechos = noechos
    _subnet = subnet
    _include_dsts = include_dsts
    _fours = fours

    poolsize = min(len(files), poolsize)
    print(poolsize)
    results = parse_parallel(files, poolsize) if poolsize != 1 else parse_sequential(files)
    if output:
        with fopen(output, 'wt') as f:
            f.writelines(a + '\n' for a in results)
    return results

def read_filemap(filename):
    filemap = {}
    with open(filename) as f:
        for line in f:
            if line[0] == '#':
                continue
            file, addr = line.split()
            filemap[file] = addr
    return filemap

def main():
    parser = ArgumentParser()
    parser.add_argument('-w', '--wfiles', help='File with list of newline-separated filenames.')
    parser.add_argument('-W', '--wfilelist', nargs='+', help='List of filenames, space separated.')
    parser.add_argument('-a', '--afiles', help='File with list of newline-separated filenames.')
    parser.add_argument('-A', '--afilelist', nargs='+', help='List of filenames, space separated.')
    parser.add_argument('-j', '--jfiles')
    parser.add_argument('-J', '--jfilelist', nargs='+')
    parser.add_argument('-p', '--poolsize', type=int, default=1)
    parser.add_argument('-o', '--output', required=True)
    parser.add_argument('-l', '--prune-loops', action='store_true')
    parser.add_argument('-e', '--noechos', action='store_true')
    parser.add_argument('-d', '--include-dsts', action='store_true')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-s', '--subnet', action='store_true')
    group.add_argument('-4', '--fours', action='store_true')
    args = parser.parse_args()
    files = []
    if args.wfiles:
        with fopen(args.wfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.WARTS) for line in f if line[0] != '#')
    if args.wfilelist:
        files.extend(TraceFile(file, OutputType.WARTS) for file in args.wfilelist)
    if args.afiles:
        with fopen(args.afiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.ATLAS) for line in f if line[0] != '#')
    if args.afilelist:
        files.extend(TraceFile(file, OutputType.ATLAS) for file in args.afilelist)
    if args.jfiles:
        with fopen(args.jfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.JSONWARTS) for line in f if line[0] != '#')
    if args.jfilelist:
        files.extend(TraceFile(file, OutputType.JSONWARTS) for file in args.jfilelist)
    ip2as = create_private()
    run(files, ip2as, args.poolsize, args.output, prune_loops=args.prune_loops, noechos=args.noechos, subnet=args.subnet, include_dsts=args.include_dsts, fours=args.fours)

if __name__ == '__main__':
    main()



from traceutils2.radix.radix cimport Radix
from traceutils2.radix.radix_node cimport RadixNodeData


cdef list PRIVATE4, PRIVATE6
cdef str MULTICAST4, MULTICAST6


cdef class IP2Data(Radix):
    cpdef RadixNodeData add_data(self, str network, short masklen=*, dict data=*);
    cpdef dict data(self, str asn);
    cpdef dict data_packed(self, bytes packed);
    cpdef void add_private(self) except *

cpdef IP2Data create_private();
cpdef IP2Data create_table(str filename);



from traceutils2.radix.radix_node cimport RadixNode
from traceutils2.radix.radix_tree cimport RadixTree
from traceutils2.radix.radix_prefix cimport RadixPrefix

cdef list _search_covering(RadixNode node);

cdef class Radix:
    cdef RadixTree _tree4
    cdef RadixTree _tree6
    cdef long gen_id

    cdef RadixNode _add(self, RadixPrefix prefix);
    cdef void _delete(self, RadixNode node) except *;
    cdef RadixNode _search_exact(self, RadixPrefix prefix);
    cdef RadixNode _search_best(self, RadixPrefix prefix);
    cdef RadixNode _search_worst(self, RadixPrefix prefix);
    cdef list _search_covered(self, RadixPrefix prefix);

    cpdef RadixNode add(self, str network, short masklen=*);
    cpdef RadixNode add_packed(self, bytes packed, unsigned char masklen);
    cpdef void delete(self, str network, short masklen=*) except *;
    cpdef void delete_packed(self, bytes packed, unsigned char masklen) except *;
    cpdef RadixNode search_exact(self, str addr);
    cpdef RadixNode search_exact_prefix(self, str network, short masklen=*);
    cpdef RadixNode search_exact_packed(self, bytes packed, unsigned char masklen);
    cpdef RadixNode search_best(self, str addr);
    cpdef RadixNode search_best_prefix(self, str network, short masklen=*);
    cpdef RadixNode search_best_packed(self, bytes packed, unsigned char masklen);
    cpdef RadixNode search_worst(self, str addr);
    cpdef RadixNode search_worst_prefix(self, str network, short masklen=*);
    cpdef RadixNode search_worst_packed(self, bytes packed, unsigned char masklen);
    cpdef list search_covered(self, str addr);
    cpdef list search_covered_prefix(self, str network, short masklen=*);
    cpdef list search_covered_packed(self, bytes packed, unsigned char masklen);
    cpdef list search_covering(self, str addr);
    cpdef list search_covering_prefix(self, str network, short masklen=*);
    cpdef list search_covering_packed(self, bytes packed, unsigned char masklen);
    cpdef list nodes(self);
    cpdef list prefixes(self)



import ujson as json
from file2 import fopen

from traceutils2.scamper.hop cimport Hop, Trace, Reader


class AtlasErrorException(Exception):
    pass

cdef list create_hops(list hops, int family):
    cdef list hopslist = []
    cdef dict h, result
    cdef list results
    cdef int i
    cdef AtlasHop hop
    if hops is not None:
        for i in range(len(hops)):
            h = hops[i]
            results = h.get('result')
            if not results:
                continue
            # print(h['result'])
            for result in results:
                if 'from' in result and 'late' not in result:
                    hop = AtlasHop(hop=h['hop'], family=family, **result)
                    hopslist.append(hop)
                    break
    return hopslist

cdef class AtlasHop(Hop):
    def __init__(self, int hop=-1, double rtt=float('nan'), int size=-1, int ttl=-1, err=None, int itos=0, int ittl=1, str flags=None, dict icmpext=None, int late=0, int dup=0, str edst=None, list hdropts=None, family=0, **kwargs):
        cdef int icmp_type, icmp_code
        # For now, if I messed up the family, assume IPv4.
        # Will change later when I get Atlas IPv6 traceroutes to test against
        if family == 0:
            family = AF_INET
        if not err:
            if family == AF_INET:
                icmp_type = 11
                icmp_code = 0
            else:
                icmp_type = 3
                icmp_code = 0
        else:
            icmp_type = 3 if family == AF_INET else 1
            if isinstance(err, str):
                if err == 'N':
                    icmp_code = 0 if family == AF_INET else 0
                elif err == 'H':
                    icmp_code = 1 if family == AF_INET else 3
                elif err == 'A':
                    icmp_code = 13 if family == AF_INET else 5
                elif err == 'P':
                    icmp_code = 2 if family == AF_INET else 4
                elif err == 'p':
                    icmp_code = 3 if family == AF_INET else 4
            elif isinstance(err, int):
                icmp_code = err
            else:
                raise AtlasErrorException('Unknown error code: {}'.format(err))
        self.addr = kwargs.get('from')
        self.probe_ttl = hop
        self.rtt = rtt
        self.reply_ttl = ttl
        self.reply_size = size
        self.icmp_type = icmp_type
        self.icmp_code = icmp_code
        self.icmpext = icmpext
        self.ismpls = bool(icmpext)
        self.reply_tos = itos
        self.icmp_q_ttl = ittl

        self.flags = flags
        self.late = late
        self.dup = dup
        self.edst = edst
        self.hdropts = hdropts

cdef class AtlasTrace(Trace):
    def __init__(self, int af=0, str dst_addr='', str dst_name='', long endtime=0, str proto='', list result=None, str src_addr='', long timestamp=0, str type='', str jdata=None, **kwargs):
        self.src = src_addr
        self.dst = dst_addr
        self.family = AF_INET if af == 4 else AF_INET6
        self.hops = create_hops(result, self.family)
        self.jdata = jdata

        self.af = af
        self.dst_name = dst_name
        self.endtime = endtime
        self.proto = proto
        self.result = result
        self.timestamp = timestamp
        self.type = type


cdef class AtlasReader(Reader):
    def __init__(self, str filename):
        self.filename = filename
        self.f = None

    def __iter__(self):
        cdef str line
        cdef dict result
        for line in self.f:
            j = json.loads(line)
            if isinstance(j, list):
                for result in j:
                    if result['type'] == 'traceroute':
                        yield AtlasTrace(jdata=line, **result)
            else:
                result = j
                if result['type'] == 'traceroute':
                    yield AtlasTrace(jdata=line, **result)

    cpdef void open(self) except *:
        self.f = fopen(self.filename)
        self.f.open()

    cpdef void close(self) except *:
        self.f.close()



import sys
from setuptools import setup, find_packages
from setuptools.extension import Extension

if 'build_ext' in sys.argv:
    from Cython.Distutils import build_ext
    use_cython = True
else:
    use_cython = False

ext_pyx = '.pyx' if use_cython else '.c'
extensions_names = {
    'bdrmapit.graph.node': ['bdrmapit/graph/node' + ext_pyx],
    'bdrmapit.graph.construct': ['bdrmapit/graph/construct' + ext_pyx],
    'bdrmapit.algorithm.updates_dict': ['bdrmapit/algorithm/updates_dict' + ext_pyx],
}

extensions = [Extension(k, v) for k, v in extensions_names.items()]
package_data = {k: ['*.pxd', '*pyx', '*.py'] for k in extensions_names}

if use_cython:
    from Cython.Build import cythonize
    extensions = cythonize(
        extensions,
        compiler_directives={'language_level': '3', 'embedsignature': True},
        annotate=True
    )

setup(
    name="bdrmapit",
    version='1.0.9',
    packages=find_packages(),
    install_requires=['jsonschema', 'traceutils2>=1.0.0', 'pandas', 'pb-amarder', 'file2', 'deprecated'],
    python_requires='>=3.6',
    ext_modules=extensions,
    entry_points={
        'console_scripts': [
            'bdrmapit=bdrmapit.scripts.bdrmapit:main',
            'traceparser=bdrmapit.scripts.traceparser:main',
            'bm_addr=bdrmapit.scripts.bm_addr:main',
            'bm_adj=bdrmapit.scripts.bm_adj:main'
        ],
    },
    zip_safe=False,
    package_data=package_data,
    include_package_data=True
)



cdef class CustomDict(dict):
    def __getstate__(self):
        return dict(self)

    def __setstate__(self, state):
        self.update(state)

    def __reduce__(self):
        return self.__class__, (), self.__getstate__()


cdef class StrDict(CustomDict):
    def __missing__(self, int key):
        return str(key)


cdef class EmptyDict(CustomDict):
    def __missing__(self, key):
        return set()


cdef class ZeroDict(CustomDict):
    def __missing__(self, key):
        return 0



from traceutils2.radix.radix cimport Radix
from traceutils2.radix.radix_node cimport RadixNodeASNs


cdef list PRIVATE4, PRIVATE6
cdef str MULTICAST4, MULTICAST6


cdef class IP2ASes(Radix):
    cpdef RadixNodeASNs add_asns(self, str network, short masklen=*, list asns=*);
    cpdef list asns(self, str asn);
    cpdef list asns_packed(self, bytes packed);
    cpdef void add_private(self) except *

cpdef IP2ASes create_private();
cpdef IP2ASes create_table(str filename);



cimport cython
from libc.string cimport strchr
from libc.string cimport memcpy
# from libc.math cimport pow

@cython.initializedcheck(False)
cdef bytes inet_pton4(bytes a):
    cdef unsigned char c[4]
    inet_pton(AF_INET, a, c)
    return <bytes>c[:4]


@cython.initializedcheck(False)
cdef bytes inet_pton6(bytes a):
    cdef unsigned char c[16]
    inet_pton(AF_INET6, a, c)
    return <bytes>c[:16]


cdef bytes inet_pton_bytes(unsigned char family, bytes a):
    if family == AF_INET:
        return inet_pton4(a)
    return inet_pton6(a)


cdef bytes inet_pton_str(unsigned char family, str s):
    cdef bytes a = s.encode()
    return inet_pton_bytes(family, a)


cdef bytes inet_pton_auto(bytes a):
    cdef int family = find_family(a)
    return inet_pton_bytes(family, a)


cdef bytes inet_pton_auto_str(str s):
    cdef bytes a = s.encode()
    return inet_pton_auto(a)


cpdef unsigned char find_family(bytes addr):
    if strchr(addr, b':'):
        return AF_INET6
    return AF_INET


@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void fix4(bytes a, unsigned char masklen, unsigned char *c) except *:
    cdef unsigned char i, quotient, remainder, mask
    quotient = masklen / 8
    remainder = masklen % 8
    inet_pton(AF_INET, a, c)
    for i in range(quotient+1, 4, 1):
        c[i] = 0
    # if remainder < 8:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask


@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void fix6(bytes a, unsigned char masklen, unsigned char *c) except *:
    cdef unsigned char i, quotient, remainder, mask
    quotient = masklen / 8
    remainder = masklen % 8
    inet_pton(AF_INET6, a, c)
    for i in range(quotient+1, 16, 1):
        c[i] = 0
    # if remainder > 0:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask


cpdef bytes inet_fix(unsigned char family, bytes a, unsigned char masklen):
    cdef unsigned char c[16]
    if family == AF_INET:
        fix4(a, masklen, c)
        return <bytes>c[:4]
    fix6(a, masklen, c)
    return <bytes>c[:16]


cdef list prefix_addrs4(bytes addr, int size):
    cdef list addrs = []
    cdef unsigned char fixed[4]
    cdef bytes newaddr_b
    cdef str newaddr
    cdef char dst[16]
    cdef unsigned char old, i
    fix4(addr, 32 - size, fixed)
    old = fixed[3]
    for i in range(2 ** size):
        fixed[3] = old + i
        inet_ntop(AF_INET, fixed, dst, INET_ADDRSTRLEN)
        newaddr_b = dst
        newaddr = newaddr_b.decode()
        addrs.append(newaddr)
    return addrs


cdef list prefix_addrs6(bytes addr, int size):
    cdef list addrs = []
    cdef unsigned char fixed[16]
    cdef bytes newaddr_b
    cdef str newaddr
    cdef char dst[46]
    cdef unsigned char old, i
    fix6(addr, 128 - size, fixed)
    old = fixed[15]
    for i in range(size ** 2):
        fixed[15] = old + i
        inet_ntop(AF_INET6, fixed, dst, INET6_ADDRSTRLEN)
        newaddr_b = dst
        newaddr = newaddr_b.decode()
        addrs.append(newaddr)
    return addrs


cpdef list prefix_addrs(str addr, int size):
    cdef bytes addr_b = addr.encode()
    cdef unsigned char family = find_family(addr_b)
    if family == AF_INET:
        return prefix_addrs4(addr_b, size)
    return prefix_addrs6(addr_b, size)


cdef bytes otherside4(bytes addr, int num):
    cdef unsigned char c[4]
    cdef bytes result
    cdef char dst[46]
    inet_pton(AF_INET, addr, c)
    if num == 2:
        if c[3] % 2 == 0:
            c[3] += 1
        else:
            c[3] -= 1
    elif num == 4:
        if c[3] % 4 == 1:
            c[3] += 1
        elif c[3] % 4 == 2:
            c[3] -= 1
        else:
            raise Exception('Invalid host address {} for /30 prefix'.format(addr.decode()))
    else:
        raise Exception('Invalid number of addresses in prefix {}'.format(num))
    inet_ntop(AF_INET, c, dst, INET_ADDRSTRLEN)
    result = <bytes>dst
    return result


cdef bytes otherside6(bytes addr, int num):
    cdef unsigned char c[16]
    cdef char dst[46]
    inet_pton(AF_INET6, addr, c)
    if num == 2:
        if c[15] % 2 == 0:
            c[15] += 1
        else:
            c[15] -= 1
    elif num == 4:
        if c[15] % 4 == 1:
            c[15] += 1
        elif c[15] % 4 == 2:
            c[15] -= 1
        else:
            raise Exception('Invalid host address {} for /126 prefix'.format(addr.decode()))
    else:
        raise Exception('Invalid number of addresses in prefix {}'.format(num))
    inet_ntop(AF_INET6, c, dst, INET6_ADDRSTRLEN)
    return <bytes>dst


cpdef str otherside(str addr, int num):
    cdef bytes addr_b = addr.encode(), result
    cdef unsigned char family = find_family(addr_b)
    if family == AF_INET:
        result = otherside4(addr_b, num)
    else:
        result = otherside6(addr_b, num)
    return result.decode()

cpdef bint valid(long asn) except -1:
    return asn != 23456 and 0 < asn < 64496 or 131071 < asn < 400000

@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void fix4_bytes(bytes a, unsigned char masklen, unsigned char *c) except *:
    cdef unsigned char i, quotient, remainder, mask
    # memcpy(c, <char *>a, 4)
    memcpy(c, <char *> a, 4)
    quotient = masklen / 8
    remainder = masklen % 8
    # inet_pton(AF_INET, a, c)
    for i in range(quotient+1, 4, 1):
        c[i] = 0
    # if remainder < 8:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask


@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void fix6_bytes(bytes a, unsigned char masklen, unsigned char *c) except *:
    cdef unsigned char i, quotient, remainder, mask
    memcpy(c, <char *>a, 16)
    quotient = masklen / 8
    remainder = masklen % 8
    # inet_pton(AF_INET6, a, c)
    for i in range(quotient+1, 16, 1):
        c[i] = 0
    # if remainder > 0:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask


cpdef bytes inet_fix_bytes(unsigned char family, bytes a, unsigned char masklen):
    cdef unsigned char c[16]
    if family == AF_INET:
        fix4_bytes(a, masklen, c)
        return <bytes>c[:4]
    fix6_bytes(a, masklen, c)
    return <bytes>c[:16]


from traceutils2.scamper.warts cimport WartsReader, WartsJsonReader
from traceutils2.scamper.atlas cimport AtlasReader

cpdef Reader reader(str filename, TraceFType ftype=TraceFType.none, bint safe=True, bint parallel_read=False):
    if ftype == TraceFType.none:
        if filename.endswith('warts') or filename.endswith('warts.gz') or filename.endswith('warts.bz2'):
            return WartsReader(filename, safe=safe)
        else:
            return WartsJsonReader(filename, safe=safe, parallel_read=parallel_read)
    if ftype == TraceFType.warts:
        return WartsReader(filename, safe=safe)
    elif ftype == TraceFType.wartsjson:
        return WartsJsonReader(filename, safe=safe, parallel_read=parallel_read)
    elif ftype == TraceFType.atlas:
        return AtlasReader(filename)
    else:
        raise Exception('Unknown file type: {}'.format(ftype))


from bdrmapit.graph.node cimport Node


cdef class UpdateObj:
    def __eq__(self, other):
        if isinstance(other, UpdateObj):
            return self.asn == other.asn
        return False

    def __repr__(self):
        return '<ASN={}, Org={}, UType={}>'.format(self.asn, self.org, self.utype)


cdef class Updates(dict):

    def __init__(self, *args, str name=None, **kargs):
        super().__init__(*args, **kargs)
        self.name = name
        self.changes = {}

    # def __setitem__(self, key, value):
    #     if self[key] != value:
    #         self.changes[key] = value

    def __missing__(self, Node key):
        return None

    cpdef void add_update(self, Node node, int asn, str org, int utype) except *:
        cdef UpdateObj update = UpdateObj()
        update.asn = asn
        update.org = org
        update.utype = utype
        if self[node] != update:
            self.changes[node] = update

    cpdef void add_update_direct(self, Node node, int asn, str org, int utype) except *:
        cdef UpdateObj update = UpdateObj()
        update.asn = asn
        update.org = org
        update.utype = utype
        self[node] = update

    cpdef void advance(self) except *:
        self.update(self.changes)
        self.changes = {}

    cpdef int asn(self, node) except *:
        cdef UpdateObj value = self[node]
        if value is not None:
            return value.asn
        return -1

    cpdef Updates make_copy(self, str name=None):
        if name is None:
            name = self.name
        return Updates(self, name=name)

    cpdef str org(self, Node node):
        cdef UpdateObj value = self[node]
        return value.org


cdef class UpdatesView(Updates):

    def __init__(self, original: Updates, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.original = original

    def __missing__(self, key):
        return self.original[key]
